"""
RT method module.
"""

# Authors: Shota Fukuda <st_fukuda@outlook.jp>
# License: BSD-3-Clause

import numpy as np
from sklearn.base import BaseEstimator
from sklearn.metrics import roc_auc_score
from sklearn.utils.validation import check_is_fitted


class RT(BaseEstimator):
    def __init__(
        self,
        *,
        ddof: int = 1,
        esp: float = 1e-16,
        threshold: float = 4.0,
        return_sqrt: bool = False,
    ):
        """
        RT method.

        The RT method is an unsupervised learning method used for pattern
        recognition in quality engineering. The method learns the mean value of
        each feature in the unit space, the sensitivity and SN ratio of each
        sample, and the associated covariance matrix of the sensitivity and SN
        ratio, and compute MD values based on these values.

        Parameters
        ----------
        ddof : int, default=1
            It means the delta degree of freedom. The divisor used in the
            calculation is `N - ddof`, where `N` is the number of samples.

        esp : float, default=1e-16
            A constant to prevent zero division. In the calculation, it is used
            as `1 / (x + esp)`.

        threshold : float, default=4.0
            Threshold. A multiple of the standard deviation of the MD values in
            the unit space. IF 4, threshold is 4 sigma.

        return_sqrt : bool, default=False
            Return the square root of the MD values or not.

        Attributes
        ----------
        mean_X_ : ndarray of shape (n_features, )
            Mean values of training data's each feature.

        mean_Y_ : ndarray of shape (2, )
            Mean values of sensitivity and error variance reciprocal.
            `mean_Y_[0]` is sensitivity mean value, and `mean_Y_[1]` is error
            variance reciprocal.

        covariance_ : ndarray of shape (2, 2)
            Variance-covariance matrix of sensitivity and error variance
            reciprocal.

        precision_ : ndarray of shape (2, 2)
            Adjoint matrix of `covariance_`.

        dist_ : ndarray of shape(n_samples, )
            Mahalanobis distances of the training set (on which fit is called)
            observations.

        n_features_in_ : int
            Number of features seen during fit.

        feature_names_in_ : ndarray of shape (n_features_in_, )
            Names of features seen during fit. Defined only when X has feature
            names that are all strings.
        """
        self.ddof = ddof
        self.esp = esp
        self.threshold = threshold
        self.return_sqrt = return_sqrt

    def fit(self, X, y=None):
        """
        Fit the model.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Training data.

        y : None
            Ignored.

        Returns
        -------
        self : object
            Fitted model.
        """
        X = self._validate_data(
            X=X,
            reset=True,
            ensure_min_samples=2,
            ensure_min_features=2,
            estimator=self,
        )

        self.mean_X_ = np.mean(X, axis=0)

        Y = self._compute_Y(X, self.mean_X_)

        self.mean_Y_ = np.mean(Y, axis=0)

        std_Y = Y - self.mean_Y_[None, :]

        self.covariance_ = np.cov(std_Y, rowvar=False, ddof=self.ddof)

        self.precision_ = self._get_precision(self.covariance_)

        self.dist_ = self._mahalanobis(Y, self.mean_Y_, self.precision_)

        self.sigma_ = np.sqrt(np.mean(self.dist_))

        return self

    def predict(self, X, y=None):
        """
        Predict labels of X according to fitted model.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Samples.

        y : None
            Ignored.

        Returns
        -------
        labels : ndarray of shape (n_samples, )
            Returns 1 for anomalies/outliers and 0 for inliers.
        """
        check_is_fitted(self)

        threshold = self.threshold * self.sigma_

        return np.where(self.mahalanobis(X=X) >= threshold, 1, 0)

    def fit_predict(self, X, y=None):
        """
        Perform fit on X and returns labels for X.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Input data.

        y : None
            Ignored.

        Returns
        -------
        labels : ndarray of shape (n_samples, )
            Returns 1 for anomalies/outliers and 0 for inliers.
        """
        return self.fit(X).predict(X)

    def mahalanobis(self, X):
        """
        Compute the mahalanobis distances (MD values).

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Samples.

        Returns
        -------
        MD : ndarray of shape (n_samples, )
            Mahalanobis distances (MD values).
        """
        check_is_fitted(self)

        X = self._validate_data(X=X, reset=False)

        Y = self._compute_Y(X, self.mean_X_)

        MD = self._mahalanobis(Y, self.mean_Y_, self.precision_)

        if self.return_sqrt:
            MD = np.sqrt(MD)

        return MD

    def score(self, X, y):
        """
        Return the ROCAUC on the given test data and labels.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Test samples.

        y : ndarray of shape (n_samples, )
            True labels for X. 1 for anomalies/outliers and 0 for inliers.

        Returns
        -------
        score : float
            ROCAUC.
        """
        check_is_fitted(self)

        X, y = self._validate_data(X=X, y=y, reset=False)

        return roc_auc_score(y, self.mahalanobis(X=X))

    def score_samples(self, X):
        """
        Compute the mahalanobis distandes (MD values).

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Samples.

        Returns
        -------
        MD : ndarray of shape (n_samples, )
            MD values.
        """
        check_is_fitted(self)

        return self.mahalanobis(X=X)

    def _compute_Y(self, X, mean_X):
        """
        Compute the sensitivity and SN ratio as Y.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Samples.

        mean_X : ndarray of shape (n_features, )
            Mean values of the unit space.

        Returns
        -------
        Y : ndarray of shape (n_samples, 2)
            Sensitivity and SN ratio. Where Y[i, 0] is the sensitivity of the
            i-th sample and Y[i, 1] is the SN ratio of the i-th sample.
        """
        std_X = X - mean_X[None, :]

        r = np.sum(mean_X**2)
        L = np.dot(std_X, mean_X)
        st = np.sum(std_X**2, axis=1)
        sb = (L**2) / (r + self.esp)
        se = st - sb
        ve = se / (std_X.shape[1] - 1)
        y1 = L / (r + self.esp)
        y2 = np.sqrt(ve)
        Y = np.hstack([y1[:, None], y2[:, None]])

        return Y

    def _get_precision(self, covariance):
        """
        Compute precision from covariance.

        Parameters
        ----------
        covariance : ndarray of shape (n_features, n_features)
            Covariance matorix of the unit space.

        Returns
        -------
        precision : ndarray of shape (n_features, n_features)
            Adjoint matrix of covariance.
        """
        precision = np.empty((2, 2))
        precision[0, 0] = covariance[1, 1]
        precision[0, 1] = -covariance[1, 0]
        precision[1, 0] = -covariance[0, 1]
        precision[1, 1] = covariance[0, 0]

        return precision

    def _mahalanobis(self, Y, mean_Y, precision):
        """
        Computes MD values.

        Parameters
        ----------
        Y : ndarray of shape (n_samples, 2)
            Sensitivity and SN ratio of samples.

        mean_Y : ndarray of shape (2, )
            Mean values of Y in the unit space.

        precision : ndarray of shape (2, 2), default=None
            Adjoint matrix of covariance in the unti space.

        Returns
        -------
        MD : ndarray of shape (n_samples, )
            Mahalanobis distances (MD values).
        """
        std_Y = Y - mean_Y[None, :]

        MD = (std_Y.dot(precision) * std_Y).mean(axis=1)

        if self.return_sqrt:
            MD = np.sqrt(MD)

        return MD

    def _more_tags(self):
        return {"binary_only": True}



"""
T(1), T(2), Ta and Tb methods module.
"""

# Authors: Shota Fukuda <st_fukuda@outlook.jp>
# License: BSD-3-Clause

import numpy as np
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils.validation import check_is_fitted


class T(RegressorMixin, BaseEstimator):
    def __init__(
        self, *, tb: bool = False, esp: float = 1e-16, is_simplified: bool = False
    ):
        """
        T(1), T(2), Ta and Tb methods.

        The T(1), T(2), Ta and Tb methods are supervised learning methods used
        for regression in quality engineering. The T(1) and T(2) methods divide
        the training data into unit space and signal data, and learn the mean
        value from the unit space and the sensitivity and SN ratio from the
        signal data. The Ta method does not divide the training data into unit
        space and signal data, and learns the mean value, sensitivity, and SN
        ratio from the all training data. The Tb method also learns from all
        training data, but for each item, the sample with the largest SN ratio
        is used as the mean value.

        Parameters
        ----------
        tb : bool, default=False
            Whether to computes as Tb method. If False, compute as T(1), T(s)
            and Ta methods.

        esp : float, default=1e-16
            A constant to prevent zero division. In the calculation, it is used
            as `1 / (x + esp)`.

        is_simplified : bool, default=False
            Compute the SN ratio with the simplified formula or not. The
            simplified formula computes with `b**2 / ve`.

        Attributes
        ----------
        mean_X_ : ndarray of shape (n_features, )
            Mean values of training data's each feature.

        mean_y_ : float or ndarray of shape (n_features, )
            Mean value of target values.

        n_ : ndarray of shape (n_features, )
            SN ratio between each feature and the target values.

        b_ : ndarray of shape (n_features, )
            Sensitivity between each feature and target values.

        n_features_in_ : int
            Number of features seen during fit.

        feature_names_in_ : ndarray of shape (n_features_in_, )
            Names of features seen during fit. Defined only when X has feature
            names that are all strings.
        """
        self.tb = tb
        self.esp = esp
        self.is_simplified = is_simplified

    def fit(self, X, y, *, us_idx=None):
        """
        Fit the model.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Training data. Includes unit space and signal data.

        y : ndarray of shape (n_samples, )
            Target values. Will be cast to X's dtype if necessary.

        us_idx : array_like of shape (n_samples, ) or None, default=None.
            A binary array indicating which sample of the training data is the
            unit space (0 for the unit space, 1 for the signal data); if None,
            the training data is not divided into the unit space and the signal
            data, but is computed as the Ta method. It is ignored when computing
            as the Tb method.

        Returns
        -------
        self : object
            Fitted model.
        """
        X, y = self._validate_data(
            X=X,
            y=y,
            reset=True,
            y_numeric=True,
            estimator=self,
        )

        if self.tb:
            n = np.empty_like(X)
            b = np.empty_like(X)
            for i, (x_i, y_i) in enumerate(zip(X, y)):
                std_X = X - x_i[None, :]
                std_y = y - y_i

                n[i], b[i] = self._compute_sn_ratio_and_sensitivity(std_X, std_y)

            idx_row = np.argmax(n, axis=0)
            idx_col = np.arange(X.shape[1])

            self.mean_X_ = X[idx_row, idx_col]
            self.mean_y_ = y[idx_row]

            self.b_ = b[idx_row, idx_col]
            self.n_ = n[idx_row, idx_col]
        else:
            if us_idx is None:
                self.mean_X_ = np.mean(X, axis=0)
                self.mean_y_ = np.mean(y)

                std_X = X - self.mean_X_[None, :]
                std_y = y - self.mean_y_
            else:
                unit_space_mask = np.where(us_idx == 0, True, False)

                self.mean_X_ = np.mean(X[unit_space_mask], axis=0)
                self.mean_y_ = np.mean(y[unit_space_mask])

                std_X = X[~unit_space_mask] - self.mean_X_[None, :]
                std_y = X[~unit_space_mask] - self.mean_y_

            self.n_, self.b_ = self._compute_sn_ratio_and_sensitivity(std_X, std_y)

        return self

    def predict(self, X, y=None):
        """
        Precit using fitted model.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Samples.

        y : None
            Ignored.

        Returns
        -------
        y_pred : ndarray of shape (n_samples, )
            Predict values.
        """
        check_is_fitted(self)

        X = self._validate_data(X=X, reset=False, estimator=self)

        std_X = X - self.mean_X_

        M_pred = std_X / (self.b_ + self.esp)[None, :]

        if self.tb:
            y_pred = M_pred + self.mean_y_[None, :]  # type: ignore
            y_pred = np.dot(y_pred, self.n_) / (np.sum(self.n_) + self.esp)
        else:
            M_pred = np.dot(M_pred, self.n_) / (np.sum(self.n_) + self.esp)
            y_pred = M_pred + self.mean_y_

        return y_pred

    def score(self, X, y):
        """
        Return the SN ratio of overall estimate.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Test samples.

        y : ndarray of shape (n_samples, )
            True values for X.

        Returns
        -------
        n : float
            SN ratio of overall estimation. For the T(1), T(2), and Ta methods,
            it is computed from M_True and M_Pred, and for the Tb method, it is
            computed from y_True and y_Pred.
        """
        check_is_fitted(self)

        X, y = self._validate_data(X=X, y=y, reset=False)

        if self.tb:
            M_true = y
            M_pred = self.predict(X)
        else:
            M_true = y - self.mean_y_
            M_pred = self.predict(X) - self.mean_y_

        r = np.sum(M_true**2)
        L = np.dot(M_true, M_pred)
        st = np.sum(M_pred**2, axis=0)
        sb = (L**2) / (r + self.esp)
        se = st - sb
        ve = se / (M_true.shape[0] - 1)

        n = (1 / r * (sb - ve)) / (ve + self.esp)
        n = 10 * np.log10(n)

        return n

    def _compute_sn_ratio_and_sensitivity(self, X, y):
        """
        Compute the SN ratio and sensitivity.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Samples.

        y : ndarray of shape (n_samples, )
            Target values.

        Returns
        -------
        n : ndarray of shape (n_features, )
            _description_

        b : ndarray of shape (n_features, )
        """
        r = np.sum(y**2)
        L = np.dot(X.T, y)
        st = np.sum(X**2, axis=0)
        sb = (L**2) / (r + self.esp)
        se = st - sb
        ve = se / (X.shape[0] - 1)

        b = L / (r + self.esp)

        if self.is_simplified:
            n = (b**2) / (ve + self.esp)
        else:
            n = (1 / r * (sb - ve)) / (ve + self.esp)
            mask = sb <= ve
            n[mask] = 0

        return n, b

    def _more_tags(self):
        return RegressorMixin._more_tags(self)



"""
MT, MTA and Standardized-Variation-Pressure methods module.
"""

# Authors: Shota Fukuda <st_fukuda@outlook.jp>
# License: BSD-3-Clause

import numpy as np
from scipy.linalg import pinvh
from scipy.stats import chi2, f
from sklearn.base import BaseEstimator
from sklearn.metrics import roc_auc_score
from sklearn.utils.validation import check_is_fitted


class MT(BaseEstimator):
    def __init__(
        self,
        *,
        method: str = "mt",
        ddof: int = 1,
        esp: float = 1e-16,
        kind: str = "specify",
        a: float = 0.05,
        threshold: float = 4.0,
        return_sqrt: bool = False,
    ):
        """
        MT, MTA and Standardized-Variation-Pressure methods.

        The MT, MTA and SVP methods are unsupervised learning methods used for
        pattern recognition in quality engineering. These methods learn the mean
        and standard deviation of each feature and the correlation matrix
        inverse of the training data, and compute MD values based on these
        values. The training data is called the unit space and usually contains
        only normal data. The MTA method learns an adjoint matrix instead of an
        inverse matrix to deal with multicolinearity. The SVP method does not
        require a correlation matrix.

        Parameters
        ----------
        method : {"mt", "mta", "svp"}, default="mt"
            Computation method.

        ddof : int, default=1
            It means the delta degree of freedom. The divisor used in the
            calculation is `N - ddof`, where `N` is the number of samples.

        esp : float, default=1e-16
            A constant to prevent zero division. In the calculation, it is used
            as `1 / (x + esp)`.

        kind : {"p", "f", "chi2", "specify"}, default="specify"
            The distribution used to determine normal and abnormal thresholds.

        a : float, default=0.05
            Right side significance level. Use to determine the threshold, if
            kind is set to f or chi2.

        threshold : float, default=4.0
            Threshold to use if kind is set to specify.

        return_sqrt : bool, default=False
            Return the square root of the MD value or not.

        Attributes
        ----------
        mean_ : ndarray of shape (n_features, )
            Mean values of each feature of the training data.

        scale_ : ndarray of shape (n_features, )
            Standard deviation values of each feature of the training data.

        covariance_ : ndarray of shape (n_features, n_features)
            Correlation matrix of the training data.

        precision_ : ndarray of shape (n_features, n_features)
            The inverse or adjoint matrix of covariance_.

        dist_ : ndarray of shape(n_samples, )
            Mahalanobis distances of the training set (on which fit is called)
            observations.

        n_features_in_ : int
            Number of features seen during fit.

        feature_names_in_ : ndarray of shape (n_features_in_, )
            Names of features seen during fit. Defined only when X has feature
            names that are all strings.
        """
        self.method = method
        self.ddof = ddof
        self.esp = esp
        self.kind = kind
        self.a = a
        self.threshold = threshold
        self.return_sqrt = return_sqrt

    def fit(self, X, y=None):
        """
        Fit the model.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Training data.

        y : None
            Ignore

        Returns
        -------
        self : object
            Fitted model.
        """
        X = self._validate_data(
            X=X,
            reset=True,
            ensure_min_samples=2,
            ensure_min_features=2,
            estimator=self,
        )

        n, p = X.shape  # type: ignore

        self.mean_ = np.mean(X, axis=0)
        self.scale_ = np.std(X, ddof=self.ddof, axis=0)

        std_X = (X - self.mean_[None, :]) / (self.scale_[None, :] + self.esp)

        self.covariance_ = np.corrcoef(std_X, rowvar=False)
        self.precision_ = self._get_precision(self.covariance_)

        self.dist_ = self._mahalanobis(X, self.mean_, self.scale_, self.precision_)

        if self.kind == "p":
            self.threshold_ = 4 * p
        elif self.kind == "f":
            self.threshold_ = (
                (p * (n - 1) * (n + 1)) / (n * (n - p)) * f.isf(self.a, p, n - p)
            )
        elif self.kind == "chi2":
            self.threshold_ = chi2.isf(self.a, p)
        else:
            self.threshold_ = self.threshold

        return self

    def predict(self, X, y=None):
        """
        Predict labels of X according to fitted model.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Samples.

        y : None
            Ignored.

        Returns
        -------
        labels : ndarray of shape (n_samples, )
            Returns 1 for anomalies/outliers and 0 for inliers.
        """
        check_is_fitted(self)

        return np.where(self.mahalanobis(X=X) >= self.threshold_, 1, 0)

    def fit_predict(self, X, y=None):
        """
        Perform fit on X and returns labels for X.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Input data.

        y : None
            Ignored.

        Returns
        -------
        labels : ndarray of shape (n_samples, )
            Returns 1 for anomalies/outliers and 0 for inliers.
        """
        return self.fit(X).predict(X)

    def mahalanobis(self, X):
        """
        Compute the mahalanobis distances (MD values).

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Samples.

        Returns
        -------
        MD : ndarray of shape (n_samples, )
            Mahalanobis distances (MD values).
        """
        check_is_fitted(self)

        X = self._validate_data(X=X, reset=False)

        MD = self._mahalanobis(X, self.mean_, self.scale_, self.precision_)

        return MD

    def score(self, X, y):
        """
        Return the ROCAUC on the given test data and labels.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Test samples.

        y : ndarray of shape (n_samples, )
            True labels for X. 1 for anomalies/outliers and 0 for inliers.

        Returns
        -------
        score : float
            ROCAUC.
        """
        check_is_fitted(self)

        X, y = self._validate_data(X=X, y=y, reset=False)

        return roc_auc_score(y, self.mahalanobis(X=X))

    def score_samples(self, X):
        """
        Compute the mahalanobis distandes (MD values).

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Samples.

        Returns
        -------
        MD : ndarray of shape (n_samples, )
            MD values.
        """
        check_is_fitted(self)

        return self.mahalanobis(X=X)

    def _cofactor(self, X: np.ndarray, i: int, j: int):
        """
        Compute the (i, j) cofactor of X.

        Parameters
        ----------
        X : ndarray of shape (n_features, n_features)
            Covariance matrix.

        i : int
            Cofactor row number. Start of 0.

        j : int
            Cofactor column number. Start of 0.

        Returns
        -------
        cofactor : float
            (i, j) cofactor of X.
        """
        M = np.delete(X, i, axis=0)
        M = np.delete(M, j, axis=1)

        return (-1) ** (i + j) * np.linalg.det(M)

    def _get_precision(self, covariance):
        """
        Compute precision from covariance.

        Parameters
        ----------
        covariance : ndarray of shape (n_features, n_features)
            Covariance matorix of the unit space.

        Returns
        -------
        precision : ndarray of shape (n_features, n_features)
            The inverse or adjoint matrix of covariance.
        """
        if self.method == "mt":
            precision = pinvh(covariance, check_finite=False)
        elif self.method == "mta":
            m = covariance.shape[0]

            precision = np.empty_like(covariance)

            for i in range(m):
                for j in range(m):
                    precision[i, j] = self._cofactor(covariance, i, j)

            precision = precision.T
        else:
            n, m = covariance.shape
            precision = np.eye(n, m)

        return precision

    def _mahalanobis(self, X, mean, scale, precision):
        """
        Compute mahalanobis distances (MD values).

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Samples.

        mean : ndarray of shape (n_features, )
            Mean values of the unit space.

        scale : ndarray of shape (n_features, )
            Standard deviation of the unit space.

        precision : ndarray of shape (n_features, n_features)
            Correlation matrix inverse of the unit space.

        Returns
        -------
        MD : ndarray of shape (n_samples, )
            Mahalanobis distances (MD values).
        """
        std_X = (X - mean[None, :]) / (scale[None, :] + self.esp)

        MD = (std_X.dot(precision) * std_X).mean(axis=1)

        if self.return_sqrt:
            MD = np.sqrt(MD)

        return MD

    def _more_tags(self) -> dict:
        return {"binary_only": True}
